# tag: ewok utility subroutines module

package Ewok::Util;

require Exporter;
use HTML::EmbperlObject;
use Fcntl qw(:flock);
use Data::Dumper;
use URI::Escape;
use File::Basename;
use Safe;
use Carp;
use Apache::Log;
use Ewok::Config qw(
    $EWOK_HOME @EWOK_STAGE_DIRS_RX @EWOK_PUBL_DIRS_RX
    %EWOK_TOP_DIRS_RESTRICT %EWOK_TOP_DIRS_RESTRICT_RX
    %EWOK_TOP_DIRS_BLOCK %EWOK_TOP_DIRS_BLOCK_RX
);

@ISA = qw(Exporter);
@EXPORT = qw();
@EXPORT_OK = qw(
    &ewokdir
    &webfsdir
    &dirCheck
    &searchpath
    &queryStrip
    &deepcopy
    &hashInvert
    &fileLoad
    &fileSave
    &executeCheckpoint
    &executeRestore
    &msCleanup
    $ERROR
    $WARNING
    $NOTICE
    $INFO
    $DEBUG
    &elog
    &tainted
);
%EXPORT_TAGS = (
    all => [@EXPORT_OK],
    checkpoint => [qw(&executeCheckpoint &executeRestore)],
    load => [qw(&fileLoad &fileSave)],
    log  => [qw($ERROR $WARNING $NOTICE $INFO $DEBUG &elog)],
);

use strict;
no strict "refs";

my %ewokdir_cache = ();

#
# Returns 'stage' if $fs_dir is in @EWOK_STAGE_DIRS_RX, 'publish' if in 
#  @EWOK_PUBLISH_DIRS_RX, else undef.
#
sub ewokdir
{
    my ($fs_dir, $arg) = @_;
    my $type = undef;

    # Return cached version, if exists
    if (exists $ewokdir_cache{$fs_dir}) {
        return $ewokdir_cache{$fs_dir};
    }

    FIND: {{
        for my $dir (@EWOK_STAGE_DIRS_RX) {
            if ($fs_dir =~ $dir) {
                $type = 'stage';
                last FIND;
            }
        }
        for my $dir (@EWOK_PUBL_DIRS_RX) {
            if ($fs_dir =~ $dir) {
                $type = 'publish';
                last FIND;
            }
        }
    }}

    # Cache
    $ewokdir_cache{$fs_dir} = $type;

    return $type;
}

#
# Checks $req->{user} and $web_dir against the Ewok::Config user-tree maps in 
#   %EWOK_TOP_DIRS_RESTRICT and %EWOK_TOP_DIRS_BLOCK. Returns 1 for access 
#   permitted, 0 for access denied, dies on error.
#
sub dirCheck
{
    my ($req, $web_dir, $arg) = @_;
    my $debug = $arg->{debug};

    # Check %EWOK_TOP_DIRS_RESTRICT
    my $restrict;
    unless ($restrict = $EWOK_TOP_DIRS_RESTRICT{$req->{user}}) {
        for my $r (keys %EWOK_TOP_DIRS_RESTRICT_RX) {
            if ($req->{user} =~ m/$r/) {
                $restrict = $EWOK_TOP_DIRS_RESTRICT_RX{$r};
                last;
            }
        }
    }
    # Return 1 if $web_dir matches one of $dir entries in $restrict, 0 otherwise
    if ($restrict) {
        die "[dirCheck] invalid %EWOK_TOP_DIRS_RESTRICT entry for '$req->{user}'"
            if ref $restrict ne 'ARRAY';
        for my $dir (@$restrict) {
            $dir =~ s!^\s*/+!!;
            return 1 if $web_dir =~ m!^/$dir/!;
        }
        return 0;
    }

    # Check %EWOK_TOP_DIRS_BLOCK
    my $block;
    unless ($block = $EWOK_TOP_DIRS_BLOCK{$req->{user}}) {
        for my $b (keys %EWOK_TOP_DIRS_BLOCK_RX) {
            if ($req->{user} =~ m/$b/) {
                $block = $EWOK_TOP_DIRS_BLOCK_RX{$b};
                last;
            }
        }
    }
    # Return 1 if $web_dir matches one of $dir entries in $block, 0 otherwise
    if ($block) {
        die "[dirCheck] invalid %EWOK_TOP_DIRS_BLOCK entry for '$req->{user}'"
            if ref $block ne 'ARRAY';
        for my $dir (@$block) {
            $dir =~ s!^\s*/+!!;
            return 0 if $web_dir =~ m!^/$dir/!;
        }
    }

    return 1;
}

#
# Derive web- and file-relative directories from the given web-relative path.
#
sub webfsdir 
{
    my ($req, $web_file, $arg) = @_;
    die "[webfsdir] no apache request found" if ! ref $req->{req_rec};
    croak "[webfsdir] invalid errors parameter" 
        if $arg->{errors} && ref $arg->{errors} ne 'ARRAY';
    
    # Cleanup and untaint $web_file
    $web_file ||= '/';
    $web_file = uri_unescape($web_file);
    ($web_file) = ($web_file =~ m:^[^/]*(.*?)\s*$:);
    $web_file =~ s!//+!/!g;          # Collapse multiple slashes
    $web_file =~ s!^/+\.+/+!/!g;     # Remove any leading . sections e.g. /./../.../foo
    $web_file =~ s![^/]+/\.\./!!g;   # Remove foo/../ sequences

    # Lookup $fs_file
    my $fs_file = $req->{req_rec}->lookup_uri($web_file)->filename();

    # Setup return values
    my ($web_dir, $fs_dir, $wd, $filename);
    if (-d $fs_file) {
        $fs_dir = $fs_file;
        $web_dir = $web_file;
        $filename = '';
    }
    else {
        ($filename, $fs_dir) = fileparse($fs_file);
        $web_dir = dirname($web_file);
    }
    $fs_dir .= '/' if substr($fs_dir,-1) ne '/';
    $web_dir .= '/' if substr($web_dir,-1) ne '/';
    $wd = $web_dir;
    $wd = substr($wd,0,-1) if $wd ne '/';

    # dirCheck $web_dir
    if (! $arg->{nocheck} && $web_dir ne '/' && ! dirCheck($req, $web_dir)) {
        push @{$arg->{errors}}, "Error: you do not have permissions for '$web_file'."
            if $arg && ref $arg->{errors} eq 'ARRAY';
        return undef;
    }
    # Check for existence
    if (! -e $fs_file) {
        push @{$arg->{errors}}, "Error: '$web_file' not found."
            if $arg && ref $arg->{errors} eq 'ARRAY';
        return undef;
    }

    return ($web_dir, $fs_dir, $wd, $filename);
}

#
# Searches the current EmbperlObject path for the given item, returning the full path to
#   the item if found, false if not found, and undef on error.
#
sub searchpath
{
    my ($req, $file) = @_;
    my $path = $req->Path;
    return undef if ! $path || ! $file;

    $path =~ s/^[:;]?(.*?)[:;]?$/$1/;
    my @path = split /[:;]+/, $path;
    for my $dir (@path) {
        return "$dir/$file" if -f "$dir/$file" && -r "$dir/$file";
    }

    return 0;
}

#
# Mimics an EmbperlObject search by searching for $file in all directories from $dir up 
#   to $stopdir, plus all directories in $addpath. Returns the full path to the item
#   if found, false if not found, and croaks on error.
#
sub searchfrom
{
    my ($dir, $file, $arg) = @_;
    my $stopdir = $arg->{stopdir};
    my $addpath = $arg->{addpath};
    croak "[searchfrom] missing argument" if ! $dir || ! $file;
    croak "[searchfrom] $dir is not a directory" if ! -d $dir;

    # Setup search path from $dir and $stopdir
    $dir =~ s:^[^/]*(.*?)/*\s*$:$1:;
    $stopdir =~ s:^[^/]*(.*?)/*\s*$:$1: if $stopdir;
    my @path = ();
    while ($dir) {
        push @path, $dir;
        last if $dir eq $stopdir;
        $dir =~ s:/+[^/]+$::;
    }
    # Add $addpath items to @path
    push @path, split(/[:;]+/, $addpath);

    # Do the actual search
    for my $dir (@path) {
        return "$dir/$file" if -f "$dir/$file" && -r "$dir/$file";
    }

    return 0;
}

# checkpoint stack, but should only ever contain ONE checkpoint
my @checkpoint = ();

#
# Save \%ENV and any hashrefs passed in (getting clobbered on publishing).
#
sub executeCheckpoint
{
#   elog($_[0], "[executeCheckpoint], ".($#_+1)." args", $INFO) 
#       if $_[0]->{req_rec};
    push @_, \%ENV;
    my $chk = [];
    for my $ref (@_) {
        my $save = {};
        map { $save->{$_} = $ref->{$_} } keys %$ref;
        push @$chk, $save;
    }
    $ENV{EXECUTE_CHECKPOINT} = 1;
    push @checkpoint, $chk;
}

#
# Restore \%ENV and any hashrefs passed in (getting clobbered on publishing).
# Should be called only after a prior executeCheckpoint and with the same args.
#
sub executeRestore
{
#   elog($_[0], "[executeRestore], ".($#_+1)." args", $INFO) 
#       if $_[0]->{req_rec};
    push @_, \%ENV;
    my $chk = shift @checkpoint;
    for my $ref (@_) {
        my $save = shift @$chk;
        # Delete any extra fields in $ref but not in $save
        map { delete $ref->{"$_"} } grep { my $z = $_; $z if ! grep /^$z$/, keys(%$save) } keys %$ref;
        map { $ref->{$_} = $save->{$_} } keys %$save;
    }
}

#
# Return the given query string $qs stripped of the variables in the $var
#   arrayref. Ensures the returned query string begins with a '&' if not null.
#
sub queryStrip 
{
    my ($req, $qs, $var) = @_;
    return '' if ! $qs;
    $var = [ split /[\s,]/, $var ] if ! ref $var;
    return '' if ! @$var;
    
    # Remove the given variables
    for my $v (@$var) {
        $qs =~ s/\b$v=[^=&]*//i;
    }
    # Compress &s
    $qs =~ s/&+/&/o;
    $qs =~ s/&$//o;
    $qs = "&$qs" if $qs ne '' && substr($qs,0,1) ne '&';

    return $qs;
}

#
# Deep copy routine, swiped from a Randal Schwartz column
#
sub deepcopy 
{
    my $this = shift;
    if (not ref $this) {
        $this;
    } elsif (ref $this eq "ARRAY") {
        [map deepcopy($_), @$this];
    } elsif (ref $this eq "HASH") {
        {map { $_ => deepcopy($this->{$_}) } keys %$this};
    } else { 
        die "what type is $_?" 
    }
}

#
# Inverts the given hash i.e. turns the keys into values and the values into
#   keys. Splits values on the given delimiter (commas and spaces by default).
#
sub hashInvert ($;$)
{
    my $hashref = shift;
    my $delim = shift;
    $delim ||= '\s*[,\s]\s*';

    my %inverse = ();
    my @values;
    for my $key (keys %$hashref) {
        for my $value (split /$delim/o, $$hashref{$key}) {
            $inverse{$value} .= ' ' if exists $inverse{$value};
            $inverse{$value} .= $key;
        }
    }

    return \%inverse;
}

#
# fileLoad will cache results if requested ('cache' => 1). Cache is keyed
#   by "$dir/$file", pointing to an arrayref of [ mtime, $serial, $eval ]
#
my %fileCache = ();

#
# Load $file, EmbperlObject-searching from $context directory, and return the contents.
#   Croaks on error or file not found.
#
sub fileLoad
{
    my ($file, $context, $arg) = @_;
    my $eval = $arg->{'eval'};
    my $name = $arg->{name};
    my $cache = $arg->{cache};
    my $inherit = $arg->{inherit};
    my $addpath = $arg->{addpath};
    $addpath = "$EWOK_HOME/htdocs" if ! exists $arg->{addpath};
    my $notfound = $arg->{notfound};
    my $ar = $arg->{ar};
    $context ||= $ENV{SCRIPT_FILENAME};
    croak "[fileLoad] invalid file '$file'" if ! $file || ref $file;
    croak "[fileLoad] invalid context '$context'" if ! $context || ref $context;
    $name ||= 'VAR1';

    my $dir;
    if (-d $context) {
        $dir = $context;
    }
    else {
        ($dir = $context) =~ s|/[^/]+$||o;
        croak "[fileLoad] cannot find directory '$dir'" if ! -d $dir;
    }
    $dir = substr($dir,0,-1) if substr($dir,-1) eq '/';

    # Check the cache
    my $mtime = (stat "$dir/$file")[9] if $cache && -e "$dir/$file";
    if ($cache && $fileCache{"$dir/$file"}) {
        my $entry = $fileCache{"$dir/$file"};
        # If cache entry is up to date, return
        if ($mtime == $entry->[0]) {
            return $eval ? $entry->[2] : $entry->[1];
        }
    }

    # Locate $file, searching from $dir
    my $inherited = 0;
    my $path = "$dir/$file" if -f "$dir/$file" && -r "$dir/$file";
    if (! $path) {
        $path = searchfrom($dir, $file, 
            { stopdir => $ENV{EMBPERL_OBJECT_STOPDIR}, addpath => $addpath });
        $inherited = 1 if $path;
    }

    # Read the data from the $path file
    my $serial = '';
    if ($path) {
        open FILE, "<", "$path"
            or croak "[fileLoad] opening file '$path' failed: $!";
        flock(FILE, LOCK_SH)
            or croak "[fileLoad] read-locking file '$path' failed: $!";
        {
            local $/ = undef;
            $serial = <FILE>;
        }
        close FILE
            or croak "[fileLoad] closing file '$path' failed: $!";
    }

    # Untaint
    ($serial) = ($serial =~ m/(\$$name = {.*})/s);

    # If no serial, give $notfound or error
    if (! $serial) {
        # If $notfound is set, failure is non-fatal - return the $notfound
        return $notfound if $notfound;
        # Otherwise this is a fatal error - croak
        croak "[fileLoad] file '$dir/$file' not found";
    }

    # If $eval flag set, eval the serial.
    if ($eval) {
        my $cp = Safe->new();
        croak "[fileLoad] Safe compartment initialisation failed for $name load"
            if ! ref $cp;
        $cp->permit_only(qw(:base_core :base_mem :base_orig));
        my $out = $cp->reval($serial);
        if ($out) {
            # Cache the results if we're caching and this is not inherited
            $fileCache{$file} = [ $mtime, $serial, $out ] if $mtime;
            return $out;
        }
        else {
            croak "[fileLoad] $name not set after eval of '$serial' ($@)";
        }
    }
    else {
        # Cache the results if we're caching and this is not inherited
        $fileCache{$file} = [ $mtime, $serial, undef ] if $mtime;
        return $serial;
    }
}

#
# Save $content (as $name, if given) into $dir/$file.
#   Croaks on failure, true on success.
#
sub fileSave
{
    my ($content, $file, $dir, $arg) = @_;
    my $name = [ $arg->{name} ] if $arg->{name};
    if ($dir eq '' || index($dir,'/') < 0) {
        croak "[fileSave] invalid directory '$dir'";
    }

    # Untaint
    ($file) = ($file =~ m|([^/]+)$|);

    # Make sure dir is a directory
    if (! -d $dir) {
        $dir =~ s|[^/]+$||o;
    }
    else {
        $dir .= '/' if substr($dir,-1) ne '/';
    }
    croak "[fileSave] directory '$dir' not found" if ! -d $dir;

    # Store the contents
    $file = $dir . $file;
    $Data::Dumper::Indent = 1;
    $Data::Dumper::Purity = 1;
    my $rc;
    open FILE, ">", $file
        or croak "[fileSave] opening file '$file' failed: $!";
    flock(FILE, LOCK_EX)
        or croak "[fileSave] write-locking file '$file' failed: $!";
    print FILE Data::Dumper->Dump([$content], $name)
        or croak "[fileSave] saving to file '$file' failed: $!";
    close FILE
        or croak "[fileSave] closing file '$file' failed: $!";
    chmod 0600, $file;

    return 1;
}

#
# Aggressively cleanup Microsoft HTML ugliness 
#
sub msCleanup ($)
{
    my ($content) = @_;
     
    # <html> attributes
    $content =~ s|<html.*?>|<html>|ois;
    # <head> stylesheets
#   $content =~ s|(<head>).*?(</head>)||ois;
    # style attributes
    $content =~ s|\s*style\s*=\s*["'][^"']+["']||oisg;
    # Change arial references to include helvetica + sans-serif
    $content =~ s|arial(\s*,s*sans-serif)?|arial,helvetica,sans-serif|oisg;
    # <!--[if/endif constructs
    $content =~ s|<!--\[if.*?<!\[endif\]-->||oisg;

    return $content;
}

# Log levels
use vars qw($DEBUG $INFO $NOTICE $WARNING $ERROR %dispatcher);
$DEBUG = 1;
$INFO = 2;
$NOTICE = 3;
$WARNING = 4;
$ERROR = 5;

%dispatcher = (
    $DEBUG   => 'debug',
    $INFO    => 'info',
    $NOTICE  => 'notice',
    $WARNING => 'warn',
    $ERROR   => 'error',
);

#
# Logging routine - log the given message to the apache error_log 
#   at the given level. $level defaults to $WARNING.
#
sub elog ($$;$$)
{
    my ($req, $msg, $level, $errors) = @_;
    $level ||= $WARNING;
    return if ! exists $dispatcher{$level};
    return if ! ref $req || ! $req->{req_rec};
    
    my $rlog = $req->{req_rec}->log;
    return if ! $rlog;

    my $method = $dispatcher{$level};
    $rlog->$method("$req->{user} $msg");

    push @$errors, "Error: $msg" if ref $errors eq 'ARRAY';
}

#
# Straight out of the Camel book III (p. 561)
#
sub tainted
{
    my $arg = shift;
    my $nada = substr($arg,0,0);
    local $@;
    eval { eval "# $nada " };
    return length($@) != 0;
}

1;

